#include <stdio.h>
#include "uart_buff.h"
#include <stdlib.h>

#define UART_MAX_BUFF_SIZE 2048
#define UART_MAX_PACKET_SIZE 256

static volatile unsigned char buff[UART_MAX_BUFF_SIZE];
static size_t start = UART_MAX_BUFF_SIZE;
static size_t packet_data_length = 0;
static volatile size_t end = 0;
static int packet_start_found = 0;
static int packet_size_found = 0;
static int packet_ready = 0;

/**
 * Put a byte into the buffer.
 */
void uart_buff_put_byte(u8 c) {
  if (uart_buff_full()) {
    return;
  }

  buff[end] = c;
  end = uart_buff_calc_index(end + 1);
}

/**
 * Scan for a packet, updating any necessary indices and flags.
 */
void uart_buff_scan_packet() {
  if (!packet_start_found) {
    uart_buff_scan_packet_start();
  }

  if (!packet_size_found) {
    uart_buff_scan_packet_size();
  }

  if (packet_start_found
      && packet_size_found
      && (uart_buff_size() >= uart_buff_packet_size())) {
    packet_ready = 1;
  }
}

/**
 * Scan for the packet start. Advance the start index of the buffer until
 * a packet start is found.
 */
void uart_buff_scan_packet_start() {
  while (buff[start] != 0xBE && !uart_buff_empty()) {
    start = uart_buff_calc_index(start + 1);
  }
  if (buff[start] == 0xBE) {
    packet_start_found = 1;
  }
}

/**
 * Scan for the packet size inside the packet.
 */
void uart_buff_scan_packet_size() {
  if (uart_buff_size() < 7) {
    // haven't received the "length" bytes yet
    return;
  }

  packet_data_length = (uart_buff_get_u8(6) << 8) | uart_buff_get_u8(5);
  packet_size_found = 1;

  if (uart_buff_packet_size() > UART_MAX_PACKET_SIZE) {
    // Packet size is too big. Abort this packet.
    uart_buff_reset_flags();
    start += 1;
  }
}

/**
 * Return 1 if a packet is ready to processed, 0 otherwise.
 */
int uart_buff_packet_ready() {
  if (!packet_ready) {
    uart_buff_scan_packet();
  }
  return packet_ready;
}

/**
 * Retrieve a 8-bit from the buffer according to the given offset with respect
 * to the start index of the buffer.
 */
u8 uart_buff_get_u8(size_t offset) {
  size_t target = uart_buff_calc_index(start + offset);

  return buff[target];
}

/**
 * Retrieve a 16-bit from the buffer according to the given offset with respect
 * to the start index of the buffer.
 */
u16 uart_buff_get_u16(size_t offset) {
  return uart_buff_data_get_u8(offset + 1) << 8 |
    uart_buff_data_get_u8(offset);
}

/**
 * Retrieve a 32-bit from the buffer according to the given offset with respect
 * to the start index of the buffer.
 */
u32 uart_buff_get_u32(size_t offset) {
  return uart_buff_data_get_u8(offset + 3) << 24 |
    uart_buff_data_get_u8(offset + 2) << 16 |
    uart_buff_data_get_u8(offset + 1) << 8 |
    uart_buff_data_get_u8(offset);
}

/**
 * Retrieve a 8-bit unsigned integer from the buffer according to
 * the given offset with respect to the start of the data portion
 * of the current packet.
 *
 * This has undefined behavior if a packet is not ready.
 */
u8 uart_buff_data_get_u8(size_t offset) {
  return uart_buff_get_u8(7 + offset);
}

/**
 * Retrieve a 16-bit unsigned integer from the buffer according to
 * the given offset with respect to the start of the data portion
 * of the current packet.
 *
 * This has undefined behavior if a packet is not ready.
 */
u16 uart_buff_data_get_u16(size_t offset) {
  return uart_buff_get_u16(7 + offset);
}

/**
 * Retrieve a 32-bit unsigned integer from the buffer according to
 * the given offset with respect to the start of the data portion
 * of the current packet.
 *
 * This has undefined behavior if a packet is not ready.
 */
u32 uart_buff_data_get_u32(size_t offset) {
  return uart_buff_get_u32(7 + offset);
}

/**
 * Retrieve a 32-bit floating point number from the buffer according to the
 * given offset with respect to the start of the data portion of the current 
 * packet.
 *
 * This has undefined behavior if a packet is not ready.
 */
float uart_buff_data_get_float(size_t offset) {
	union {
		float f;
		int i;
	} x;
	x.i = uart_buff_data_get_u8(offset + 3) << 24 |
			uart_buff_data_get_u8(offset + 2) << 16 |
			uart_buff_data_get_u8(offset + 1) << 8 |
			uart_buff_data_get_u8(offset);
	return x.f;
}

/**
 * Move the start index of the buffer to the end of the current packet.
 * If a packet is not ready, this function does nothing.
 */
void uart_buff_flush_packet() {
  if (!packet_ready) {
    return;
  }

  start = uart_buff_calc_index(start + uart_buff_packet_size() + 1);
  uart_buff_reset_flags();
}

/**
 * Return the current number of bytes held in the buffer.
 */
size_t uart_buff_size() {
  return uart_buff_calc_index(end - start);
}

/**
 * Return the length of the data portion of the current packet.
 */
size_t uart_buff_data_length() {
  return packet_data_length;
}

/**
 * Return the size of the current packet.
 */
size_t uart_buff_packet_size() {
  return 1 + 1 + 1 + 2 + packet_data_length + 1;
}

/**
 * Reset all flags indicating the NOTEstatus of scanned packets.
 */
void uart_buff_reset_flags() {
  packet_start_found = 0;
  packet_size_found = 0;
  packet_ready = 0;
}

/**
 * Return 1 if the buffer is empty, 0 otherwise.
 */
int uart_buff_empty() {
  int effective_end = uart_buff_calc_index(end - 1);
  return start == effective_end;
}

/**
 * Return 1 if the buffer is full, 0 otherwise.
 */
int uart_buff_full() {
  return start == end;
}

/**
 * Calculate the actual index from the given index, wrapping around
 * the edges of the circular buffer if necessary.
 */
size_t uart_buff_calc_index(int given_index) {
  int actual_index = given_index;
  if (actual_index < 0) {
    actual_index += UART_MAX_BUFF_SIZE;
  }
  if (actual_index >= UART_MAX_BUFF_SIZE) {
    actual_index -= UART_MAX_BUFF_SIZE;
  }
  return actual_index;
}

/**
 * Return a string pointer to the start of the data section.
 *
 * ALERT! Consumers of this function must free the pointer they received,
 * or else they will introduce a memory leak!
 */
char * uart_buff_get_packet() {
	size_t packet_size = uart_buff_packet_size();
	char *str = malloc(sizeof(char) * (packet_size + 1));
	str[packet_size + 1] = '\0'; // null terminate string
	int i;
	for (i = 0; i < packet_size; i += 1) {
		str[i] = uart_buff_get_u8(i);
	}
	return str;
}

/**
 * Print the buffer with a pretty ASCII art image.
 */
void uart_buff_print() {
  size_t packet_size = uart_buff_packet_size();
  int i;
  printf("buffer size = %zu\n", uart_buff_size());
  puts("Buffer:");
  puts(".--.");
  puts("|  |");
  puts("|  v");
  for (i = 0; i < UART_MAX_BUFF_SIZE; i += 1) {
    printf("|  %02X", buff[i]);
    if (i == start) {
      printf("<-- start");
    }
    if (packet_ready) {
      if (i == uart_buff_calc_index(start + 7)) {
        printf("<-- data start");
      }
      if (i == uart_buff_calc_index(start + packet_size)) {
        printf("<-- data end");
      }
    }
    if (i == end) {
      printf("<-- end");
    }
    printf("\n");
  }
  puts("|  |");
  puts("'--'");
}
